Introduce suspend_lock to prevent concurrent instances of
dm_swap_table, dm_suspend and dm_resume.
--- diff/drivers/md/dm.c	2005-07-06 19:08:31.000000000 +0100
+++ source/drivers/md/dm.c	2005-07-06 19:08:27.000000000 +0100
@@ -59,6 +59,7 @@
 
 struct mapped_device {
 	struct rw_semaphore io_lock;
+	struct rw_semaphore suspend_lock;
 	rwlock_t map_lock;
 	atomic_t holders;
 
@@ -748,6 +749,7 @@
 
 	memset(md, 0, sizeof(*md));
 	init_rwsem(&md->io_lock);
+	init_rwsem(&md->suspend_lock);
 	rwlock_init(&md->map_lock);
 	atomic_set(&md->holders, 1);
 	atomic_set(&md->event_nr, 0);
@@ -964,7 +966,7 @@
 {
 	int r = -EINVAL;
 
-	down_write(&md->io_lock);
+	down_write(&md->suspend_lock);
 
 	/* device must be suspended */
 	if (!dm_suspended(md))
@@ -974,7 +976,7 @@
 	r = __bind(md, table);
 
 out:
-	up_write(&md->io_lock);
+	up_write(&md->suspend_lock);
 	return r;
 }
 
@@ -1043,6 +1045,8 @@
 	DECLARE_WAITQUEUE(wait, current);
 	int r = -EINVAL;
 
+	down_write(&md->suspend_lock);
+
 	/* Flush I/O to the device. */
 	down_read(&md->io_lock);
 	if (test_bit(DMF_BLOCK_IO, &md->flags))
@@ -1119,10 +1123,12 @@
 	clear_bit(DMF_BLOCK_IO, &md->flags);
 out_write_unlock:
 	up_write(&md->io_lock);
+	up_write(&md->suspend_lock);
 	return r;
 
 out_read_unlock:
 	up_read(&md->io_lock);
+	up_write(&md->suspend_lock);
 	return r;
 }
 
@@ -1131,11 +1137,13 @@
 	struct bio *def;
 	struct dm_table *map = dm_get_table(md);
 
+	down_write(&md->suspend_lock);
 	down_write(&md->io_lock);
 	if (!map ||
 	    !dm_suspended(md) ||
 	    !dm_table_get_size(map)) {
 		up_write(&md->io_lock);
+		up_write(&md->suspend_lock);
 		dm_table_put(map);
 		return -EINVAL;
 	}
@@ -1150,6 +1158,7 @@
 	__unlock_fs(md);
 	dm_table_unplug_all(map);
 	dm_table_put(map);
+	up_write(&md->suspend_lock);
 
 	return 0;
 }
